/* MERG Content-to-template merger
   Command line:
      MERG [switches] template content1 content2 ...

   Valid switches:
      h        allow/Suppress HTML tags in content
      n        Keep/discard name
      l        long file name Acceptable/truncate
      e ERROR  Output error messages to ERROR
               instead of STD_ERR
      q        Quiet mode


tstream.c Token stream manipulation functions
This file contains a set of functions used by the parsers
in manipulating a "token stream". 
*/

#include "merg.h"

TOKENSTREAM * CreateStream()
{
	TOKENSTREAM * result;

	result = (TOKENSTREAM *)MemAlloc(sizeof(TOKENSTREAM));
	if(result == NULL)
		return NULL;
	result->count = 1;
	result->namecount = 0;
	result->offset = 0;
	result->boffset = 0;
	result->start = (TOKEN *)MemAlloc(sizeof(TOKEN));
	result->names = (NAMESPACE *)MemAlloc(sizeof(NAMESPACE));
	result->names->name = NULL;
	(result->start)->nTokenID = TKN_START;

	return result;
}

#if defined(_MERG_DUPLICATESTREAM)
TOKENSTREAM * DuplicateStream (TOKENSTREAM * which)
{
	TOKENSTREAM * duplicate;
	TOKEN * work, * src;
	char * tmp;
	int i, j;

	duplicate = (TOKENSTREAM *)MemAlloc(sizeof(TOKENSTREAM));
	if(duplicate == NULL)
		return NULL;

	duplicate->count = which->count;
	duplicate->namecount = which->namecount;
	duplicate->offset = which->offset;
	duplicate->boffset = which->boffset;
	
	i = StreamCount(which);
	duplicate->start = (TOKEN *) MemAlloc (sizeof(TOKEN) * i);
	duplicate->start = (TOKEN *) memcpy (duplicate->start, which->start, sizeof(TOKEN) * i);
	
	i = 0;
	
	/* Time to copy each string */
	for (i = 1; i < (duplicate->count); i++){
		work = GetToken(duplicate, i);
		src = GetToken(which, i);
		CreateToken (work);
		TokenContent(work, TokenContent(src, NULL));
	}
	
	if(i = which->namecount){
		duplicate->names = (NAMESPACE *) MemAlloc (sizeof(NAMESPACE) * i);
		duplicate->names = (NAMESPACE *) memcpy (duplicate->names, which->names, sizeof(NAMESPACE) * i);
		for(j=0; j < i; j++){
			(duplicate->names)[j].name = (char*) MemAlloc (sizeof(char) * strlen((which->names)[j].name));
			(duplicate->names)[j].name = strcpy((duplicate->names)[j].name, (which->names)[j].name);
		}
	}

	return duplicate;
}
#endif

inline int StreamCount (TOKENSTREAM *ts)
{
	return ts->count;
}

int FreeStream (TOKENSTREAM * ts)
{
	int i, j;

	/* First, we have to walk the stream to free the token strings, etc. */
	i = ts->count;
	while(j < i){
		FreeToken((ts->start)+j);
		j++;
	}

	/* Then we should free the name space */
	i = ts->namecount;
	while(i-- > 0)
		MemFree(((ts->names)+i)->name);

	MemFree(ts->names);
	MemFree(ts->start);
	MemFree(ts);
	return 0;
}

inline int StreamOffsetToEnd(TOKENSTREAM *ts)
{
	return (ts->offset = StreamCount(ts) - 1);
}

inline int StreamOffsetToStart(TOKENSTREAM *ts)
{
	return (ts->offset = 0);
}

int StreamOffset(TOKENSTREAM *ts, int offset)
{
	ts->offset+=offset;
	if(ts->offset >= StreamCount(ts)){
		StreamOffsetToEnd(ts);
		return -1;
	}
	if(ts->offset < 0){
		StreamOffsetToStart(ts);
		return -1;
	}

	return ts->offset;
}

int StreamOffsetAbs(TOKENSTREAM *ts, int abs)
{
	if (abs < 0)
		StreamOffsetToStart(ts);
	else if (abs >= StreamCount(ts))
		StreamOffsetToEnd(ts);
	else
		ts->offset = abs;
	return 0;
}



/* WARNING: This version of AddToken() ignores
   memory allocation problems on realloc(). Should
   be fixed on next version */

int AddToken (TOKENSTREAM * ts, TOKEN * token)
{
	int i, curcount = StreamCount (ts);
	TOKEN * tmp;

	StreamOffsetToEnd(ts);

	ts->start = (TOKEN *)MemRealloc(ts->start, sizeof(TOKEN)*(curcount+1));

	i=++(ts->offset);
	tmp = (ts->start) + i;
	CreateToken (tmp);
	TokenContent(tmp, TokenContent(token, NULL));
	TokenName(tmp, TokenName(token, NULL));
	TokenType(tmp, TokenType(token, 0));

	return ++(ts->count);
}

/* This function is currently unimplemented */
int DeleteToken (TOKENSTREAM * ts, TOKEN * token)
{
	return -1;
}

/* This function is currently unimplemented */
int DeleteTokenO (TOKENSTREAM * ts, int offset)
{
	return -1;
}

int AttachName(TOKENSTREAM * ts, int offset)
{
	int nc;
	NAMESPACE *target;
	TOKEN * token;
	char * temp;
	
	token = GetToken(ts, offset);
	
	nc=ts->namecount;
	ts->names=(NAMESPACE *)MemRealloc(ts->names, sizeof(NAMESPACE) * (nc+2));
	target = (ts->names+nc);
	temp = TokenName(token, NULL);
	target->name = (char *)MemAlloc(sizeof(char) * (strlen(temp)+1));
	strcpy(target->name, temp);
	target->offset = offset;

	ts->namecount=++nc;

	return 0;
}

int SetStreamName(TOKENSTREAM * ts, char * name)
{
	ts->streamname = name;
	return 0;
}

inline int OffsetToken (TOKENSTREAM * ts)
{
	return ts->offset;
}

TOKEN * GetToken (TOKENSTREAM * ts, int o)
{
	return ts->start+o;
}

/* This function is currently unimplemented */
int SearchToken (TOKENSTREAM * ts, TOKEN * token)
{
	return -1;
}

int MatchTokenO (TOKENSTREAM * ts, char * orig)
{
	int i = 0, j = 0, rval = -1;
	char * tmp;
	char * name = (char *) MemAlloc (sizeof(char) * strlen(orig));

	name = strcpy(name, orig);
	for(j=0; j < strlen(orig); j++)
		name[j] = tolower(name[j]);


	while(i < ts->namecount){
		tmp = (ts->names)[i].name;
		if(!strcmp(name, tmp)){
			rval = (ts->names)[i].offset;
			break;
		}
		i++;
	}
	
	MemFree (name);

	return rval;
}

TOKEN * MatchToken (TOKENSTREAM * ts, char * name)
{
	int i;
	i = MatchTokenO (ts, name);
	if(i != -1)
		return GetToken(ts, i);
	else
		return NULL;
}


/* The first call to FindTokenType searches for
the first occurence of a token type. Subsequent
calls search for the next occurences; pass 0 as
type. Returns NULL when there are no more tokens
to process */

TOKEN * FindTokenType (TOKENSTREAM * ts, int type)
{
	static int stype;
	int retval;
	TOKEN * current;

	if(type){
		stype = type;
		retval = StreamOffsetToStart(ts);
	}
	else
		retval = OffsetToken(ts);
	
	while((retval = StreamOffset(ts, 1)) != -1){
		current = GetToken(ts, retval);
		if(current->nTokenID == stype)
			return current;
	}

	return NULL;
}

inline NAMESPACE * GetNamespace (TOKENSTREAM * ts)
{
	return ts->names;
}

void WalkStream (TOKENSTREAM * ts, FILE * pipe)
{
	int i;
	TOKEN * current;
	char * temp;
	
	if(!ts)
		return;

	i = StreamCount(ts);
	current = ts->start;
	while (i > 1){		/* an ugly kludge */
		current++;
		temp = (current->szContent);
		fprintf(pipe, "%s", temp);
		i--;
	}
	return;
}

